package ca.uwaterloo.cs.jgrok.env;

import java.io.InputStream;
import java.io.PrintStream;
import java.util.ArrayList;

import ca.uwaterloo.cs.jgrok.interp.LookupException;
import ca.uwaterloo.cs.jgrok.interp.Scope;
import ca.uwaterloo.cs.jgrok.interp.ScriptUnitNode;
import ca.uwaterloo.cs.jgrok.interp.Variable;
import ca.uwaterloo.cs.jgrok.lib.Function;
import ca.uwaterloo.cs.jgrok.lib.FunctionLib;
import ca.uwaterloo.cs.jgrok.lib.FunctionLibManager;

public class Env {
    public InputStream in  = System.in;
    public PrintStream out = System.out;
    public PrintStream err = System.err;
    
    // Prompt text for JGrok interpreter.
    public static String promptText = ">> ";
    public static String JGrokHome = System.getProperty("JGrokHome");
    public static String JGrokEnvFile = System.getProperty("JGrokEnvFile");
    
    // Best guess!
    static {
        if(JGrokHome == null) {
            JGrokHome = ".";
        }
        if(JGrokEnvFile == null) {
            JGrokEnvFile = "./env/MainEnv.xml";
        }
    }
    
    /**
     * A stack of scopes.
     */
    private ArrayList<Scope> scopes;
    
    /**
     * The main script unit.
     */
    private ScriptUnitNode mainUnit;
    
    /**
     * The function library manager.
     */
    private FunctionLibManager fLibManager;
    
    /**
     * The associated tracing utility. 
     */
    private Tracing trcUtility;
    
    /**
     * Creates a new JGrok environment.
     */
    public Env() {
        scopes = new ArrayList<Scope>(10);
        fLibManager = new FunctionLibManager();
        trcUtility = new Tracing();
        load(Env.JGrokEnvFile);
    }
    
    /**
     * Loads an user specified environment file.
     * @param envFileName the file to load.
     */
    public void load(String envFileName) {
        try {
            new EnvLoader().load(this, envFileName);
        } catch (LoadingException e) {
            e.printStackTrace(err);
        }
    }
    
    /**
     * Gets the tracing utility.
     */
    public Tracing getTracing() {
        return trcUtility;
    }
    
    /**
     * Gets the FunctionLib manager
     */
    public FunctionLibManager getFunctionLibManager() {
        return fLibManager;
    }
    
    /**
     * Gets the main script unit.
     */
    public ScriptUnitNode getMainUnit() {
        return mainUnit;
    }
    
    /**
     * Sets the main script unit.
     */
    public void setMainUnit(ScriptUnitNode unit) {
        mainUnit = unit;
    }
    
    /**
     * Removes all scopes on the stack.
     */
    public void removeAllScopes() {
        scopes.clear();
    }

    /**
     * Pops the scope on top of the stack.
     */
    public Scope popScope() {
        return (Scope)scopes.remove(scopes.size()-1);
    }
    
    /**
     * Pushes a scope onto the stack.
     */
    public void pushScope(Scope scp) {
        scopes.add(scp);
    }
    
    /**
     * Peeps the scope on top of the stack.
     */
    public Scope peepScope() {
        return (Scope)scopes.get(scopes.size()-1);
    }
    
    /**
     * Traces the script unit on top of the stack.
     */
    public ScriptUnitNode traceScriptUnit() {
        Scope scp = peepScope();
        if(scp instanceof ScriptUnitNode) {
            return (ScriptUnitNode)scp;
        } else {
            while(scp != null) {
                scp = scp.getParent();
                if(scp instanceof ScriptUnitNode)
                    return (ScriptUnitNode)scp;
            }
            return null;
        }
    }
    
    /**
     * Looks up a variable.
     * @throws LookupException if no variable can be found.
     */
    public Variable lookup(String name) throws LookupException {
        Variable var = null;
        LookupException lex = null;
        
        int i = scopes.size() - 1;
        while(i >= 0) {
            Scope scp = (Scope)scopes.get(i);
            
            try {
                var = scp.lookup(name);
                return var;
            } catch(LookupException e) {
                if(lex == null) lex = e;
            }
            
            i--;
        }
        
        if(lex != null) throw lex;
        else return var;
    }

    /**
     * Looks up all functions with a given name in the root FunctionLib.
     * @param functionName the name of functions to look up.
     */
    public Function[] lookupFunctions(String functionName)
        throws LookupException {
        Function[] fs;
        fs = fLibManager.getRootLib().find(functionName);
        if(fs.length != 0) return fs;
        throw new LookupException("function " + functionName);
    }
    
    /**
     * Looks up all functions with a given name in the root FunctionLib.
     * @param functionName the name of functions to look up.
     */
    public Function[] lookupFunctions(String libName, String functionName)
        throws LookupException {
        FunctionLib lib = fLibManager.findLib(libName);
        if (lib != null) {
            Function[] fs = lib.find(functionName);
            if(fs.length != 0) return fs;
        }
        throw new LookupException("function " + functionName);
    }

    /**
     * Looks up a Function in the root FunctionLib.
     * @param functionName the name of Function to look up.
     * @param paramTypes the parameter types of the Function.
     */
    public Function lookupFunction(String functionName, Class<?>[] paramTypes)
        throws LookupException {
        Function aFunction;
        aFunction = fLibManager.getRootLib().find(functionName, paramTypes);
        if(aFunction != null) return aFunction;
        throw new LookupException("function " + functionName);
    }
    
    /**
     * Looks up a Function in the root FunctionLib.
     * @param functionName the name of Function to look up.
     * @param paramTypes the parameter types of the Function.
     */
    public Function lookupFunction(String libName, String functionName, Class<?>[] paramTypes)
        throws LookupException {
        FunctionLib lib = fLibManager.findLib(libName);
        if(lib != null) {
            Function aFunction = lib.find(functionName, paramTypes);
            if(aFunction != null) return aFunction;
        }
        throw new LookupException("function " + libName + "." + functionName);
    }
    
    /**
     * Tests if a given name is JGrok reserved word.
     * A reserved word cannot be assignment left in JGrok.
     */
    public static boolean isReservedWord(String name) {
        if(name.equals("class") ||
           name.equals("JGrok")) {
            return true;
        }
        
        if(name.equals("int")   ||
           name.equals("long")  ||
           name.equals("byte")  ||
           name.equals("short") ||
           name.equals("float") ||
           name.equals("double")||
           name.equals("boolean") ) {
            return true;
        }
        
        return false;
    }
}
