package ca.uwaterloo.cs.jgrok.lib;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;

/**
 * Shows edges along the contain hierarchy.
 * 
 * <pre>
 * + Show edges inside the specified node:
 *     void showedge(string node, EdgeSet edgeSet)
 *     void showedge(string node, EdgeSet edgeSet, int level)
 *
 * + Show edges from the source node to the target node:
 *     void showedge(string srcNode, string trgNode, EdgeSet edgeSet)
 *     void showedge(string srcNode, string trgNode, EdgeSet edgeSet, int level)
 * </pre>
 */
public class ShowEdge extends Function {
    private Env env;
    private int whichMethod;
    private StringBuffer indent = new StringBuffer();
    
    public static final int METHOD_DEFAULT = 0;
    public static final int METHOD_ROOT_NODE = 1;
    public static final int METHOD_ROOT_NODE_LEVEL = 2;
    public static final int METHOD_SOURCE_DESTINATION = 3;
    public static final int METHOD_SOURCE_DESTINATION_LEVEL = 4;
    
    public ShowEdge(int whichMethod) {
        this();
        this.whichMethod = whichMethod;
    }
    
    private ShowEdge() {
        name = "showedge";
        type = void.class;
        paramTypes = new Class[0];
        whichMethod = 0;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        Tree2 tree2;
        EdgeSet contain;
        this.env = env;
        
        try {
            contain = findContain();
            tree2 = new Tree2(contain);
        } catch(InvocationException e) {
            env.out.println("trivial showedge, not implemented");
            return Value.VOID;
        }
        
        switch(whichMethod) {
        case 1:
            {
                String node = vals[0].toString();
                EdgeSet links = (EdgeSet)vals[1].objectValue();
                
                int top = IDManager.getID(node);
                Tree edgeTree = tree2.getEdgeTree(top, links);
                EdgeSet multiplicity = new EdgeSet();
                calculateMultiplicity(top, edgeTree, multiplicity, true);
                printNode(top, edgeTree, links.getName(), multiplicity);
            }
            break;
        case 2:
            {
                String node = vals[0].toString();
                EdgeSet links = (EdgeSet)vals[1].objectValue();
                int level = vals[2].intValue();

                if(level < 0)
                    throw new InvocationException("illegal showedge level: " + level);
                
                int top = IDManager.getID(node);
                Tree edgeTree = tree2.getEdgeTree(top, links);
                EdgeSet multiplicity = new EdgeSet();
                calculateMultiplicity(top, edgeTree, multiplicity, true);
                printNode(top, edgeTree, links.getName(), multiplicity, 0, level);
            }
            break;
        case 3:
            {
                String src = vals[0].toString();
                String trg = vals[1].toString();
                EdgeSet links = (EdgeSet)vals[2].objectValue();
                
                int srcID;
                int trgID;
                int edgeRoot;
                Tree edgeTree;
                
                srcID = IDManager.getID(src);
                trgID = IDManager.getID(trg);
                edgeTree = tree2.getEdgeTree(srcID, trgID, links);
                edgeRoot = edgeTree.getRoots()[0];
                
                EdgeSet multiplicity = new EdgeSet();
                calculateMultiplicity(edgeRoot, edgeTree, multiplicity, true);
                printEdge(edgeRoot, edgeTree, true, links.getName(), multiplicity);
            }
            break;
        case 4:
            {
                String src = vals[0].toString();
                String trg = vals[1].toString();
                EdgeSet links = (EdgeSet)vals[2].objectValue();
                int level = vals[3].intValue();
                
                if(level < 0)
                    throw new InvocationException("illegal showedge level: " + level);

                int srcID;
                int trgID;
                int edgeRoot;
                Tree edgeTree;
                
                srcID = IDManager.getID(src);
                trgID = IDManager.getID(trg);
                edgeTree = tree2.getEdgeTree(srcID, trgID, links);
                edgeRoot = edgeTree.getRoots()[0];
                
                EdgeSet multiplicity = new EdgeSet();
                calculateMultiplicity(edgeRoot, edgeTree, multiplicity, true);
                printEdge(edgeRoot, edgeTree, true, links.getName(), multiplicity, 0, level);
            }
            break;
        }
        
        return Value.VOID;
    }
    
    private EdgeSet findContain()
        throws InvocationException {
        try {
            Variable var;
            var = env.peepScope().lookup("contain");
            if(var.getType() != EdgeSet.class) {
                throw new InvocationException("contain is not relation");
            }
            return (EdgeSet)var.getValue().objectValue();
        } catch(Exception e) {
            throw new InvocationException("contain not found");
        }
    }
    
    private int calculateMultiplicity(int node,
                                      Tree tree,
                                      EdgeSet eSet,
                                      boolean isRoot) {
        int multiplicity = 0;
        int[] children = tree.getChildren(node);
        if(children.length == 0) {
            if(isRoot)
                multiplicity = 0;
            else
                multiplicity = 1;
        } else {
            for(int i = 0; i < children.length; i++) {
                multiplicity += calculateMultiplicity(children[i], tree, eSet, false);
            }
        }
        
        eSet.add(node, IDManager.getID(multiplicity + ""));
        return multiplicity;
    }
    
    private void printNode(int node,
                           Tree tree,
                           String relName,
                           EdgeSet multiplicity) {
        String num = ShowDB.getAtt(node, multiplicity);
        env.out.println(indent.toString()
                        + num + " : " + IDManager.get(node));
        
        int[] children;        
        children = tree.getChildren(node);
        
        if(children.length > 0) {
            incrIndent();
            for(int i = 0; i < children.length; i++) {
                printEdge(children[i], tree, false, relName, multiplicity);
            }
            decrIndent();
        }
    }
    
    private void printNode(int node,
                           Tree tree,
                           String relName,
                           EdgeSet multiplicity,
                           int curLevel, int endLevel) {
        String num = ShowDB.getAtt(node, multiplicity);
        env.out.println(indent.toString()
                        + num + " : " + IDManager.get(node));
        if(curLevel == endLevel) return;
        
        int[] children;        
        children = tree.getChildren(node);
        
        if(children.length > 0) {
            incrIndent();
            for(int i = 0; i < children.length; i++) {
                printEdge(children[i], tree, false, relName,
                          multiplicity, curLevel+1, endLevel);
            }
            decrIndent();
        }
    }
    
    private void printEdge(int edge,
                           Tree tree,
                           boolean isRoot,
                           String relName,
                           EdgeSet multiplicity) {
        String s;
        if(isRoot) s = "=>";
        else s = IDManager.get(edge);
        
        String num = ShowDB.getAtt(edge, multiplicity);
        int[] children = tree.getChildren(edge);
                
        if(children.length == 0) {
            // Get rid of '(' and ')'
            s = s.substring(1, s.length()-1);
            
            StringBuffer buffer;
            buffer = new StringBuffer();
            buffer.append(indent.toString());
            buffer.append(' ');
            buffer.append("-> (");
            buffer.append(relName);
            buffer.append(' ');
            buffer.append(s);
            buffer.append(')');
            
            env.out.println(buffer.toString());
        } else {
            if(isRoot) 
                env.out.println(indent.toString()
                                + num + " : " + s);
            else {
                int[] parse;
                parse = IDManager.parse(edge);
                if(parse.length == 1)
                    env.out.println(indent.toString()
                                    + num + " : " + s);
                else
                    env.out.println(indent.toString()
                                    + num + " : => " + s);
            }
            
            incrIndent();
            for(int i = 0; i < children.length; i++) {
                printEdge(children[i], tree, false, relName, multiplicity);
            }
            decrIndent();
        }
    }
    
    private void printEdge(int edge,
                           Tree tree,
                           boolean isRoot,
                           String relName,
                           EdgeSet multiplicity,
                           int curLevel, int endLevel) {
        String s;
        if(isRoot) s = "=>";
        else s = IDManager.get(edge);
        
        String num = ShowDB.getAtt(edge, multiplicity);
        int[] children = tree.getChildren(edge);
        
        if(children.length == 0) {
            // Get rid of '(' and ')'
            s = s.substring(1, s.length()-1);
            
            StringBuffer buffer;
            buffer = new StringBuffer();
            buffer.append(indent.toString());
            buffer.append(' ');
            buffer.append("-> (");
            buffer.append(relName);
            buffer.append(' ');
            buffer.append(s);
            buffer.append(')');
            
            env.out.println(buffer.toString());
        } else {
            if(isRoot) 
                env.out.println(indent.toString()
                                + num + " : " + s);
            else {
                int[] parse;
                parse = IDManager.parse(edge);
                if(parse.length == 1)
                    env.out.println(indent.toString()
                                    + num + " : " + s);
                else
                    env.out.println(indent.toString()
                                    + num + " : => " + s);
            }
            
            if(curLevel == endLevel) return;
            
            incrIndent();
            for(int i = 0; i < children.length; i++) {
                printEdge(children[i], tree, false, relName,
                          multiplicity, curLevel+1, endLevel);
            }
            decrIndent();
        }
    }
    
    private void incrIndent() {
        indent.append("|  ");
    }
    
    private void decrIndent() {
        int len = indent.length();
        if(len > 0) indent.delete(len-3, len);
    }
}
